;;; crc40.lisp -- a CR calculation that uses 40 bit integers
;;
;; from Raymond Toy


;(in-package :cl-bench.crc)

(declaim (optimize (debug 0)))

;#-mkcl
#+(or)
(declaim (inline crc-division-step))

;#-mkcl
#+(or)
(defun crc-division-step (bit rmdr poly msb-mask)
  (declare (type (signed-byte 56) rmdr poly msb-mask)
	   (type bit bit))
  ;; Shift in the bit into the LSB of the register (rmdr)
  (let ((new-rmdr (logior bit (the fixnum (* rmdr 2)))))
    (declare (type (signed-byte 56) new-rmdr))
    ;; Divide by the polynomial, and return the new remainder
    (if (zerop (the (signed-byte 56) (logand msb-mask new-rmdr)))
	new-rmdr
	(logxor new-rmdr poly))))

;#+mkcl
#+(and)
(defmacro crc-division-step (bit rmdr poly msb-mask)
  `(let ((bit ,bit)
	 (rmdr ,rmdr)
	 (poly ,poly)
	 (msb-mask ,msb-mask))
     (declare (type (signed-byte 56) rmdr poly msb-mask)
	      (type bit bit))
     ;; Shift in the bit into the LSB of the register (rmdr)
     (let ((new-rmdr (logior bit (the fixnum (* rmdr 2)))))
       (declare (type (signed-byte 56) new-rmdr))
       ;; Divide by the polynomial, and return the new remainder
       (the (signed-byte 56)
	 (if (zerop (the (signed-byte 56) (logand msb-mask new-rmdr)))
	     new-rmdr
	   (the (signed-byte 56) (logxor new-rmdr poly)))))))


(defun compute-adjustment (poly n)
  (declare (type (signed-byte 56) poly)
	   (fixnum n))
  ;; Precompute X^(n-1) mod poly
  (let* ((poly-len-mask (ash 1 (the fixnum (1- (integer-length poly)))))
	 (rmdr (crc-division-step 1 0 poly poly-len-mask)))
    (declare (type (signed-byte 56) poly-len-mask rmdr))
    (dotimes (k (the fixnum (- n 1)))
      (declare (fixnum k))
      (setf rmdr (crc-division-step 0 rmdr poly poly-len-mask)))
    rmdr))

(defun calculate-crc40 (iterations)
  (declare (fixnum iterations))
  (let ((crc-poly 1099587256329)
	(len 3014633)
	(answer 0))
    (dotimes (k iterations)
      (declare (fixnum k))
      (setf answer (compute-adjustment crc-poly len)))
    answer))

(defun run-crc40 ()
  (format t "~&crc40 answered: ~S.~%" (time (calculate-crc40 10))))

;; EOF
