

(defun simple-consing ()
  (loop repeat #|500000|# 500000 do
	(loop repeat 100 collect nil)))

(defvar *gate-lock* (mt:make-lock :name "start-end gate lock"))

(defvar *start-gate* (mt:make-condition-variable))
(defvar *start-gate-state* :up)
(defvar *start-gate-count* 0)
(defvar *start-gate-expected-count* -1)

(defvar *end-gate* (mt:make-condition-variable))
;(defvar *gate-lock* (mt:make-lock :name "end gate lock"))
(defvar *end-gate-state* :up)
(defvar *end-gate-count* 0)
(defvar *end-gate-expected-count* -1)

(defun init-gates (expected-count)
  (setq *start-gate-state* :up)
  (setq *start-gate-count* 0)
  (setq *end-gate-state* :up)
  (setq *end-gate-count* 0)
  (setq *start-gate-expected-count* expected-count)
  (setq *end-gate-expected-count* expected-count)
  )

(defun wait-at-start-gate ()
  (mt:with-lock (*gate-lock*)
    (format t "~&Waiting at start gate in thread ~S~%" mt:*thread*)
    (finish-output)
    (incf *start-gate-count*) ;; make yourself counted at the gate.
    (when (eql *start-gate-count* *start-gate-expected-count*)
      (setq *start-gate-state* :down)
      (mt:condition-broadcast *start-gate*)
      (format t "~&It's a go!~%") (finish-output)
      )
    (si:while (eq *start-gate-state* :up)
      (mt:condition-wait *start-gate* *gate-lock*))
    ;;(format t "~&Running in thread ~S.~%" mt:*thread*)
    (finish-output)
    )
  )

(defun wait-at-end-gate ()
  (mt:with-lock (*gate-lock*)
    (format t "~&Waiting at end gate in thread ~S~%" mt:*thread*)
    (finish-output)
    (incf *end-gate-count*) ;; make yourself counted at the gate.
    (when (eql *end-gate-count* *end-gate-expected-count*)
      (setq *end-gate-state* :down)
      (mt:condition-broadcast *end-gate*)
      (format t "~&It's a finish!~%") (finish-output)
      )
    (si:while (eq *end-gate-state* :up)
      (mt:condition-wait *end-gate* *gate-lock*))
    ;;(format t "~&Running in thread ~S.~%" mt:*thread*)
    (finish-output)
    )
  )

(defun run-func-grouped (func)
  (wait-at-start-gate)
  (funcall func)
  (wait-at-end-gate)
  :done)


(defun run-gc-scale-test (nb-threads)
  (let (threads)
    (format t "~&In run-gc-scale-test for nb-threads = ~S.~%" nb-threads)
    (init-gates (1+ nb-threads))
    (dotimes (i nb-threads)
      (push (mt:thread-run-function (format nil "Worker ~D" i) #'run-func-grouped #'simple-consing) threads)
      )
    (wait-at-start-gate)
    ;;(format t "~&Work being done!~%") (finish-output)
    (wait-at-end-gate)
    (format t "~&Work done.~%") (finish-output)
    (format t "~&Collecting results.~%") (finish-output)
    (dolist (thr threads)
      (let ((val (mt:join-thread thr)))
	(unless (eq val :done) (format t "~&Thread ~S returned abnormal value: ~S.~%" thr val))))
    (format t "~&All done.~%") (finish-output)
    )
  )

